using Adaptation.Eaf.Management.ConfigurationData.CellAutomation;
using Adaptation.FileHandlers.json.WorkItems;
using Adaptation.Ifx.Eaf.EquipmentConnector.File.Configuration;
using Adaptation.Shared;
using Adaptation.Shared.Duplicator;
using Adaptation.Shared.Methods;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text.Json;
using System.Threading;

namespace Adaptation.FileHandlers.json;

public class FileRead : Shared.FileRead, IFileRead
{

    private readonly Timer _Timer;

    public FileRead(ISMTP smtp, Dictionary<string, string> fileParameter, string cellInstanceName, int? connectionCount, string cellInstanceConnectionName, FileConnectorConfiguration fileConnectorConfiguration, string equipmentTypeName, string parameterizedModelObjectDefinitionType, IList<ModelObjectParameterDefinition> modelObjectParameters, string equipmentDictionaryName, Dictionary<string, List<long>> dummyRuns, Dictionary<long, List<string>> staticRuns, bool useCyclicalForDescription, bool isEAFHosted) :
        base(new Description(), false, smtp, fileParameter, cellInstanceName, connectionCount, cellInstanceConnectionName, fileConnectorConfiguration, equipmentTypeName, parameterizedModelObjectDefinitionType, modelObjectParameters, equipmentDictionaryName, dummyRuns, staticRuns, useCyclicalForDescription, isEAFHosted: connectionCount is null)
    {
        _MinFileLength = 10;
        _NullData = string.Empty;
        _Logistics = new(this);
        if (_FileParameter is null)
            throw new Exception(cellInstanceConnectionName);
        if (_ModelObjectParameterDefinitions is null)
            throw new Exception(cellInstanceConnectionName);
        if (_IsDuplicator)
            throw new Exception(cellInstanceConnectionName);
        if (_IsEAFHosted)
            NestExistingFiles(_FileConnectorConfiguration);
        if (!Debugger.IsAttached && fileConnectorConfiguration.PreProcessingMode != FileConnectorConfiguration.PreProcessingModeEnum.Process)
            _Timer = new Timer(Callback, null, (int)(fileConnectorConfiguration.FileScanningIntervalInSeconds * 1000), Timeout.Infinite);
        else
        {
            _Timer = new Timer(Callback, null, Timeout.Infinite, Timeout.Infinite);
            Callback(null);
        }
    }

    void IFileRead.Move(Tuple<string, Test[], JsonElement[], List<FileInfo>> extractResults, Exception exception) => Move(extractResults);

    void IFileRead.WaitForThread() => WaitForThread(thread: null, threadExceptions: null);

    string IFileRead.GetEventDescription()
    {
        string result = _Description.GetEventDescription();
        return result;
    }

    List<string> IFileRead.GetHeaderNames()
    {
        List<string> results = _Description.GetHeaderNames();
        return results;
    }

    string[] IFileRead.Move(Tuple<string, Test[], JsonElement[], List<FileInfo>> extractResults, string to, string from, string resolvedFileLocation, Exception exception)
    {
        string[] results = Move(extractResults, to, from, resolvedFileLocation, exception);
        return results;
    }

    JsonProperty[] IFileRead.GetDefault()
    {
        JsonProperty[] results = _Description.GetDefault(this, _Logistics);
        return results;
    }

    Dictionary<string, string> IFileRead.GetDisplayNamesJsonElement()
    {
        Dictionary<string, string> results = _Description.GetDisplayNamesJsonElement(this);
        return results;
    }

    List<IDescription> IFileRead.GetDescriptions(IFileRead fileRead, List<Test> tests, IProcessData processData)
    {
        List<IDescription> results = _Description.GetDescriptions(fileRead, _Logistics, tests, processData);
        return results;
    }

    Tuple<string, Test[], JsonElement[], List<FileInfo>> IFileRead.GetExtractResult(string reportFullPath, string eventName)
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        if (string.IsNullOrEmpty(eventName))
            throw new Exception();
        _ReportFullPath = reportFullPath;
        DateTime dateTime = DateTime.Now;
        results = GetExtractResult(reportFullPath, dateTime);
        if (results.Item3 is null)
            results = new Tuple<string, Test[], JsonElement[], List<FileInfo>>(results.Item1, Array.Empty<Test>(), JsonSerializer.Deserialize<JsonElement[]>("[]"), results.Item4);
        if (results.Item3.Length > 0 && _IsEAFHosted)
            WritePDSF(this, results.Item3);
        UpdateLastTicksDuration(DateTime.Now.Ticks - dateTime.Ticks);
        return results;
    }

    Tuple<string, Test[], JsonElement[], List<FileInfo>> IFileRead.ReExtract()
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        List<string> headerNames = _Description.GetHeaderNames();
        Dictionary<string, string> keyValuePairs = _Description.GetDisplayNamesJsonElement(this);
        results = ReExtract(this, headerNames, keyValuePairs);
        return results;
    }

#nullable enable

    private static ReadOnlyCollection<WorkItem> GetWorkItems(ReadOnlyCollection<Value> valueWithReqCollection)
    {
        List<WorkItem> results = new();
        Fields fields;
        WorkItem workItem;
        foreach (Value value in valueWithReqCollection)
        {
            fields = value.Fields;
            workItem = new(fields.MicrosoftVSTSCommonActivatedDate == DateTime.MinValue ? null : fields.MicrosoftVSTSCommonActivatedDate,
                           fields.SystemAreaPath,
                           fields.SystemAssignedTo?.DisplayName,
                           fields.MicrosoftVSTSCommonBusinessValue is null or 0 ? null : (long)fields.MicrosoftVSTSCommonBusinessValue,
                           fields.SystemChangedDate,
                           fields.MicrosoftVSTSCommonClosedDate == DateTime.MinValue ? null : fields.MicrosoftVSTSCommonClosedDate,
                           fields.SystemCommentCount,
                           fields.SystemCreatedDate,
                           fields.SystemDescription,
                           fields.MicrosoftVSTSSchedulingEffort is null or 0 ? null : (long)fields.MicrosoftVSTSSchedulingEffort,
                           value.Id,
                           fields.SystemIterationPath,
                           fields.SystemParent == 0 ? null : fields.SystemParent,
                           fields.MicrosoftVSTSCommonPriority == 0 ? null : fields.MicrosoftVSTSCommonPriority,
                           value.Relations,
                           fields.CustomRequester?.DisplayName,
                           fields.MicrosoftVSTSCommonResolvedDate == DateTime.MinValue ? null : fields.MicrosoftVSTSCommonResolvedDate,
                           value.Rev,
                           fields.CustomRRminusOE is null or 0 ? null : (long)fields.CustomRRminusOE,
                           fields.MicrosoftVSTSSchedulingStartDate == DateTime.MinValue ? null : fields.MicrosoftVSTSSchedulingStartDate,
                           fields.SystemState,
                           fields.SystemTags,
                           fields.MicrosoftVSTSSchedulingTargetDate == DateTime.MinValue ? null : fields.MicrosoftVSTSSchedulingTargetDate,
                           fields.MicrosoftVSTSCommonTimeCriticality is null or 0 ? null : (long)fields.MicrosoftVSTSCommonTimeCriticality,
                           fields.SystemTitle,
                           null,
                           fields.CustomWSJF is null or 0 ? null : (long)fields.CustomWSJF,
                           fields.SystemWorkItemType);
            results.Add(workItem);
        }
        return new(results);
    }

    private static ReadOnlyCollection<Value> GetWorkItems(string[] files)
    {
        List<Value> results = new();
        string json;
        Value? value;
        foreach (string file in from l in files orderby l.Length, l select l)
        {
            json = File.ReadAllText(file);
            value = JsonSerializer.Deserialize<Value>(json);
            if (value is null)
                continue;
            results.Add(value);
        }
        return new(results);
    }

    private static void WriteCollectionJson(FileConnectorConfiguration fileConnectorConfiguration, string[] alternateTargetFolders, ReadOnlyCollection<WorkItem> workItems)
    {
        string old;
        string checkFile;
        string? pathRoot;
        JsonSerializerOptions jsonSerializerOptions = new() { WriteIndented = true };
        string json = JsonSerializer.Serialize(workItems, jsonSerializerOptions);
        foreach (string alternateTargetFolder in alternateTargetFolders)
        {
            if (alternateTargetFolder == fileConnectorConfiguration.TargetFileLocation)
                continue;
            pathRoot = Path.GetPathRoot(alternateTargetFolder);
            if (string.IsNullOrEmpty(pathRoot))
                continue;
            try
            {
                if (!Directory.Exists(pathRoot))
                    continue;
            }
            catch (Exception)
            { continue; }
            if (!Directory.Exists(alternateTargetFolder))
                _ = Directory.CreateDirectory(alternateTargetFolder);
            checkFile = Path.Combine(alternateTargetFolder, fileConnectorConfiguration.TargetFileName);
            old = !File.Exists(checkFile) ? string.Empty : File.ReadAllText(checkFile);
            if (old == json)
                continue;
            File.WriteAllText(checkFile, json);
        }
    }

    private static void ParseWorkItemsAsync(FileConnectorConfiguration fileConnectorConfiguration, Calendar calendar, string[] alternateTargetFolders)
    {
        if (!Directory.Exists(fileConnectorConfiguration.TargetFileLocation))
            _ = Directory.CreateDirectory(fileConnectorConfiguration.TargetFileLocation);
        DateTime dateTime = DateTime.Now;
        string weekOfYear = calendar.GetWeekOfYear(dateTime, CalendarWeekRule.FirstDay, DayOfWeek.Sunday).ToString("00");
        string weekDirectory = Path.Combine(fileConnectorConfiguration.TargetFileLocation, $"{dateTime:yyyy}_Week_{weekOfYear}");
        string[] files = Directory.GetFiles(fileConnectorConfiguration.TargetFileLocation, fileConnectorConfiguration.SourceFileFilter, SearchOption.TopDirectoryOnly);
        if (!Directory.Exists(weekDirectory))
        {
            _ = Directory.CreateDirectory(weekDirectory);
            foreach (string file in files)
                File.Copy(file, Path.Combine(weekDirectory, Path.GetFileName(file)));
        }
        ReadOnlyCollection<Value> collection = GetWorkItems(files);
        ReadOnlyCollection<WorkItem> workItems = GetWorkItems(collection);
        if (workItems.Count > 0)
            WriteCollectionJson(fileConnectorConfiguration, alternateTargetFolders, workItems);
    }

    private static void ParseWorkItemsAsync(FileConnectorConfiguration fileConnectorConfiguration, Calendar calendar)
    {
        string[] alternateTargetFolders = fileConnectorConfiguration.AlternateTargetFolder.Split('|');
        if (alternateTargetFolders.Length > 0)
            ParseWorkItemsAsync(fileConnectorConfiguration, calendar, alternateTargetFolders);
    }

    private void Callback(object state)
    {
        try
        {
            if (_IsEAFHosted)
                ParseWorkItemsAsync(_FileConnectorConfiguration, _Calendar);
        }
        catch (Exception exception)
        {
            string subject = string.Concat("Exception:", _CellInstanceConnectionName);
            string body = string.Concat(exception.Message, Environment.NewLine, Environment.NewLine, exception.StackTrace);
            try
            { _SMTP.SendHighPriorityEmailMessage(subject, body); }
            catch (Exception) { }
        }
        try
        {
            if (_FileConnectorConfiguration?.FileScanningIntervalInSeconds is null)
                throw new Exception();
            TimeSpan timeSpan = new(DateTime.Now.AddSeconds(_FileConnectorConfiguration.FileScanningIntervalInSeconds.Value).Ticks - DateTime.Now.Ticks);
            _ = _Timer.Change((long)timeSpan.TotalMilliseconds, Timeout.Infinite);
        }
        catch (Exception exception)
        {
            string subject = string.Concat("Exception:", _CellInstanceConnectionName);
            string body = string.Concat(exception.Message, Environment.NewLine, Environment.NewLine, exception.StackTrace);
            try
            { _SMTP.SendHighPriorityEmailMessage(subject, body); }
            catch (Exception) { }
        }
    }

#pragma warning disable IDE0060
    private void MoveJson(string reportFullPath, DateTime dateTime)
#pragma warning restore IDE0060
    {
        string json = File.ReadAllText(reportFullPath);
        Value? value = JsonSerializer.Deserialize<Value>(json);
        if (value is null)
            throw new Exception(nameof(value));
        string checkFile = Path.Combine(_FileConnectorConfiguration.TargetFileLocation, $"{value.Id}.json");
        string old = !File.Exists(checkFile) ? string.Empty : File.ReadAllText(checkFile);
        if (json != old)
        {
            File.Delete(checkFile);
            File.Move(reportFullPath, checkFile);
        }
    }

    private Tuple<string, Test[], JsonElement[], List<FileInfo>> GetExtractResult(string reportFullPath, DateTime dateTime)
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        _Logistics = new Logistics(reportFullPath, $"LOGISTICS_1{'\t'}A_JOBID={"BACKLOG"};A_MES_ENTITY={"BACKLOG"};");
        if (_IsEAFHosted && _FileConnectorConfiguration.FileScanningIntervalInSeconds > 0)
            MoveJson(reportFullPath, dateTime);
        results = new(_Logistics.Logistics1[0], Array.Empty<Test>(), Array.Empty<JsonElement>(), new List<FileInfo>());
        return results;
    }

}